@@ -33,7 +33,7 @@
 
 #ifndef lint
 /*static char sccsid[] = "from: @(#)common.c	5.7 (Berkeley) 3/2/91";*/
-static char rcsid[] = "$Id: common.c,v 1.3 1999/02/16 03:45:31 notting Exp $";
+static char rcsid[] = "$Id: common.c,v 1.7 1999/10/17 20:07:12 notting Exp $";
 #endif /* not lint */
 
 /*
@@ -275,7 +275,7 @@ checkremote()
 			"unable to get official name for local machine %s",
 			name);
 		    return errbuf;
-		} else (void) strcpy(name, hp->h_name);
+		} else (void) strncpy(name, hp->h_name, MAXHOSTNAMELEN);
 
 		/* get the official name of RM */
 		hp = gethostbyname(RM);
@@ -314,16 +314,16 @@ checkfromremote(fromhost)
        name[sizeof(name)-1] = '\0';
        hp = gethostbyname(name);
        if (hp == (struct hostent *) NULL) {
-           (void) sprintf(errbuf,
+           (void) snprintf(errbuf, 128,
                "unable to get official name for local machine %s",
                name);
            return 1;
-       } else (void) strcpy(name, hp->h_name);
+       } else (void) strncpy(name, hp->h_name, MAXHOSTNAMELEN);
 
        /* get the official name of fromhost */
        hp = gethostbyname(fromhost);
        if (hp == (struct hostent *) NULL) {
-           (void) sprintf(errbuf,
+           (void) snprintf(errbuf, 128,
                "unable to get official name for remote machine %s",
                fromhost);
            return 1;
