@@ -41,6 +41,7 @@
 #include "lp.h"
 #include "pathnames.h"
 
+
 #include <string.h>
 
 #define DORETURN	0	/* absorb fork error */
@@ -77,6 +78,7 @@ static char	fromhost[MAXHOSTNAMELEN];		/* user's host machine */
 				/* indentation size in static characters */
 static char	indent[10] = "-i0";
 static char	jobname[100];		/* job or file name */
+static char	extparms[256];		/* extended parms */
 static char	length[10] = "-l";	/* page length in lines */
 static char	logname[32];		/* user's login name */
 static char	euser[32];		/* effective user */
@@ -159,7 +161,7 @@ printjob()
 	/*
 	 * write process id for others to know
 	 */
-	sprintf(line, "%u\n", pid);
+	snprintf(line, BUFSIZ, "%u\n", pid);
 	pidoff = i = strlen(line);
 	if (write(lfd, line, i) != i) {
 		syslog(LOG_ERR, "%s: %s: %m", printer, LO);
@@ -345,6 +347,22 @@ printit(file)
 	 *      getline reads a line and expands tabs to blanks
 	 */
 
+	/* pass 0 */
+    /* Henri Gomez <gomez@slib.fr> 19991104 */
+	/* Pre scan to get the Z flags (RFC Filter Extension) */
+	while (getline(cfp))
+		switch (line[0]) {
+        case 'Z':
+            if (line[1] != '\0') {
+                strncpy(extparms, line+1, sizeof(extparms) - 1);
+                extparms[sizeof(extparms) - 1] = '\0';
+            } else
+                strcpy(extparms, " ");
+            continue;
+	}
+
+	rewind(cfp);
+
 	/* pass 1 */
 
 	while (getline(cfp))
@@ -359,24 +377,25 @@ printit(file)
 			continue;
 
 		case 'P':
+			{
+			struct passwd *dude;
+
 			strncpy(logname, line+1, sizeof(logname) - 1);
 			logname[sizeof(logname) - 1] = '\0';
-			if (RS) {			/* restricted */
-				if (getpwnam(logname) == NULL) {
+			dude = getpwnam(logname);
+
+			if (RS && dude == NULL) {	/* no such user plus
+							 * restricted */
 					bombed = NOACCT;
 					sendmail(line+1, bombed);
 					goto pass2;
-				}
-			} else {			/* not restricted */
-				struct passwd *dude;
-				dude = getpwnam(logname);
-				if (dude == NULL || dude->pw_uid == 0) {
-				   /* if user dne, set euid to lp  */
-				   strncpy(euser, "lp", 3);
-				} else {
+			}
+			if(dude == NULL || dude->pw_uid == 0)
+				   /* if user root, set euid to lp  */
+				strncpy(euser, "lp", 3);
+			else
 				   /* since the user does exist run filter as user */
-				   strncpy(euser, logname, sizeof(logname));
-				}
+				strncpy(euser, logname, sizeof(euser));
 			}
 			continue;
 
@@ -393,6 +412,9 @@ printit(file)
 			fino = i;
 			continue;
 
+		case 'Z':
+			continue; /* Allready handled */
+
 		case 'J':
 			if (line[1] != '\0') {
 				strncpy(jobname, line+1, sizeof(jobname) - 1);
@@ -516,9 +538,33 @@ print(format, file)
 	int pid, p[2], stopped = 0;
 	union wait status;
 	struct stat stb;
+	struct passwd *dude;
+	uid_t euid;
+	gid_t egid;
+	
+	euid = geteuid();
+	egid = getegid();
+	
+	if ((dude=getpwnam(euser)) == (struct passwd *)0 || !*euser)
+	        return(ERROR);
+	
+	if (setegid(dude->pw_gid) < 0)
+	  return (ERROR);
+	if (seteuid(dude->pw_uid) < 0)
+	  return (ERROR);
+	
+	/* Open the file as the user. */
 
-	if (lstat(file, &stb) < 0 || (fi = open(file, O_RDONLY)) < 0)
-		return(ERROR);
+	if (lstat(file, &stb) < 0 || (fi = open(file, O_RDONLY)) < 0) {
+		seteuid(euid);
+		setegid(egid);
+		return(ACCESS);
+	}
+	
+	if (seteuid(euid) < 0 || setegid(egid) < 0) {
+		syslog(LOG_ERR, "can't reset effective uid/gid!");
+		return (ERROR);
+	}
 	/*
 	 * Check to see if data file is a symbolic link. If so, it should
 	 * still point to the same file or someone is trying to print
@@ -532,11 +578,6 @@ print(format, file)
 		tof = 1;
 	}
 
-
-	/* lets make sure we know what user to run the filters as */
-	if (getpwnam(euser) == (struct passwd *)0 || !*euser)
-	        return(ERROR);
-	
 	if (IF == NULL && (format == 'f' || format == 'l')) {
 		tof = 0;
 		while ((n = read(fi, buf, BUFSIZ)) > 0)
@@ -661,6 +702,19 @@ print(format, file)
 	av[n++] = logname;
 	av[n++] = "-h";
 	av[n++] = fromhost;
+
+	/* Add Jobname */
+	if (jobname[0] && (jobname[0] != ' ')) {
+		av[n++] = "-j";
+		av[n++] = jobname;
+	}
+
+	/* Add Extended Parms */
+	if (extparms[0] && (extparms[0] != ' ')) {
+		av[n++] = "-z";
+        av[n++] = extparms;
+	}
+
 	av[n++] = AF;
 	av[n] = 0;
 	fo = pfd;
@@ -749,6 +803,7 @@ sendit(file)
 	char *dup_cfpname = NULL;
 	char *final_cfp;
 	FILE *dup_cfp = NULL;
+	struct stat stb;
 
 	int  ifi, fi;
 	int  ifchild, pid;
@@ -756,30 +811,34 @@ sendit(file)
 	int  tmpmask;
 
 	int  ii;
-
+    
+    
 	/*
 	 * open control file
 	 */
 	if ((cfp = fopen(file, "r")) == NULL)
 		return(OK);
 
-	/*
-         * if IF exists, we need to mirror the cfp file
-         */
-	if (IF != NULL)
-	  {
-	    dup_cfpname = strdup(file);
-	    memcpy((dup_cfpname+7), "XXXXXX", 6);
-	    mktemp(dup_cfpname);
-	    tmpmask = umask(007);
-	    dup_cfp=fopen(dup_cfpname, "w");
-	    umask(tmpmask);
-	    if (dup_cfp == NULL)
-	      {
+	/* Don't duplicate the cfpname, becuase if it doesn't
+	 * get deleted, weird stuff will happen. */
+	dup_cfpname=strdup("lpdtempXXXXXX");
+	if ((ii=mkstemp(dup_cfpname))==-1) {
+		syslog(LOG_ERR, "Can't create temp cfp file");
+		return(OK);
+	}
+	close(ii);
+	tmpmask = umask(007);
+	if (rename(file,dup_cfpname)==-1) {
+		syslog(LOG_ERR, "Can't rename cfp file");
+		return(OK);
+	}
+	dup_cfp=fopen(file, "w");
+	umask(tmpmask);
+	if (dup_cfp == NULL)
+	{
 		syslog(LOG_ERR, "Can't create temp cfp file %s", dup_cfpname);
 		return(OK);
-	      }
-	  }
+	}
 	    
 	/*
 	 *      read the control file for work to do
@@ -840,35 +899,95 @@ sendit(file)
 
 		/* a hack */
 		if (line[0] < 'a' || line[0] > 'z') {
-		  if (dup_cfp && line[0] != 'U')
-		    fprintf(dup_cfp,"%s\n",line);
+			
+		  if (dup_cfp && line[0] != 'U') {
+			  if (!SH) {
+				  fprintf(dup_cfp,"%s\n",line);
+			  } else if (line[0] != 'J' && line[0] != 'C' && line[0] != 'L') {
+				  fprintf(dup_cfp,"%s\n",line);
+			  }
+		  }
 		  continue;
 		}
 		  
 		/* we assume that we've hit a line dedicated to printing */
 		if (line[0] >= 'a' && line[0] <= 'z') {
-			strcpy(last, line);
-			while (i = getline(cfp))
+		        char tmpfile[20];
+			
+			memset(tmpfile,'\0',20);
+			if (dup_cfp)
+			  fprintf(dup_cfp, "%s\n", line);
+
+		        strcpy(last, line);
+			/* handle multiple copies */
+                       while (i = getline(cfp)) {
+                         if (dup_cfp)
+                           fprintf(dup_cfp, "%s\n", line);
 				if (strcmp(last, line))
 					break;
+                       }
+			
 
 			/* msf - here is where we hack in IF support    */
 			/*       will only run the IF, no other filters */
 			/*       will be run.                           */
+                       /* This hack will work for multiple files only  */
+                       /* if the 'f' lines are separated by at least   */
+                       /* one non-'f' line (i.e. a U or N line) due to */
+                       /* the multiple copies getline() while loop     */
+                       /* right above this comment.                    */
 			if (IF != NULL && last[0] == 'f') {
-			    char tmpfile[20];
+			    struct passwd *dude;
+			    uid_t euid;
+			    gid_t egid;
 			  
 			    memcpy(tmpfile, last+1, 7);
   			    memcpy(tmpfile+7, "XXXXXX", 6);
 			    tmpfile[13]='\000';
 			    mktemp(tmpfile);
 
-			    /* we make a new f and U entry for remote cfp */
-			    fprintf(dup_cfp, "f%s\n",tmpfile);
-			    fprintf(dup_cfp, "U%s\n",tmpfile);
+			    /* Nasty trick of the day, to attempt
+			     * to be RFC compliant */
+			    if (rename(last+1,tmpfile)==-1) {
+				    syslog(LOG_ERR,"can't rename data file");
+				    return(ERROR);
+			    }
+				
+			    euid = geteuid();
+			    egid = getegid();
+	
+			    /* Make sure we open the file as the user */
+			    if ((dude=getpwnam(euser)) == (struct passwd *)0 || !*euser) {
+				    syslog(LOG_ERR,"don't know who to print as!");
+				    return(ERROR);
+			    }
+	
+			    if (setegid(dude->pw_gid) < 0)
+				    return(ERROR);
+			    if (seteuid(dude->pw_uid) < 0)
+				    return(ERROR);
 
-			    fi = open(last+1, O_RDONLY);
-			    ifi = open(tmpfile,O_WRONLY|O_CREAT|O_TRUNC, 0664);
+			     fi = open(tmpfile, O_RDONLY);
+				
+			     /* Ideally here, if the open fails, we'd
+			      * just bail. However, doing this instead
+			      * allows us to clean up the queue nicely. 
+			      */
+			     if (fi==-1 ||
+				 ((stb.st_mode & S_IFMT) == S_IFLNK 
+				  && fstat(fi, &stb) == 0 &&
+				  (stb.st_dev != fdev || stb.st_ino != fino))
+				 ) {
+				     syslog(LOG_ERR, "%s: couldn't open spool file %s as %s",
+					    name, last+1, euser);
+				     fi = open("/dev/null",O_RDONLY);
+			     }
+			     if (seteuid(euid) < 0 || setegid(egid) < 0) {
+				     syslog(LOG_ERR, "%s: couldn't reset effective uid/gid!\n", name);
+				     return(ERROR);
+			     }
+				
+			    ifi = open(last+1,O_WRONLY|O_CREAT|O_TRUNC, 0664);
 			    if ((ifchild = doforkuser(DORETURN,euser)) == 0) { /*child*/
 			         dup2(fi,  0); /* input file is stdin  */
   			         dup2(ifi, 1); /* temp file is  stdout */
@@ -882,23 +1001,22 @@ sendit(file)
 			  close(ifi);
                           close(fi);
 			  /* wait on it to finish */
+			  signal(SIGCHLD,SIG_DFL);
 			  waitpid(ifchild, NULL, 0);
-			      
-			  strcpy(last+1, tmpfile);
 			 }
-                        else
-    		         {
-			  if (dup_cfp)
-			    fputs(line, dup_cfp);
-                         }
 		      
                         sendresult = sendfile('\3', last+1);
 
-    		        if ( IF != NULL )
+    		        if ( IF != NULL && last[0] == 'f')
 			  {
-			     if (unlink(last+1) != 0)
-			     syslog(LOG_ERR, "error unlinking %s: %d",
-				   last+1,errno);
+			     if (unlink(last+1) != 0) {
+				     syslog(LOG_ERR, "error unlinking %s: %d",
+					    last+1,errno);
+			     }
+			     if (rename(tmpfile,last+1)==-1) {
+				     syslog(LOG_ERR, "error renaming %s: %d",
+					    tmpfile, errno);
+			     }
 			  }
 
 			/* clean up our temp file */
@@ -921,19 +1039,18 @@ sendit(file)
 
 	if (dup_cfp)
 	  {
-	    final_cfp=dup_cfpname;
 	    fclose(dup_cfp);
 	    dup_cfp = NULL;
 	  }
-	else
-	  final_cfp=file;
+        final_cfp=file;
 
 	if (err == OK && sendfile('\2', final_cfp) > 0) {
 		(void) fclose(cfp);
 		if (dup_cfp)
 		  {
 		    (void) fclose(dup_cfp);
-		    unlink(dup_cfpname);
+		    unlink(file);
+		    rename(dup_cfpname,file);
 		    free(dup_cfpname);
 		  }
 		return(REPRINT);
@@ -968,7 +1085,7 @@ pass2:
 	 * clean-up in case another control file exists
 	 */
 	(void) fclose(cfp);
-	(void) unlink(file);
+	unlink(file);
 	if (dup_cfpname) {
 	    unlink(dup_cfpname);
 	    free(dup_cfpname);
@@ -990,9 +1107,36 @@ sendfile(type, file)
 	struct stat stb;
 	char buf[BUFSIZ];
 	int sizerr, resp;
+	struct passwd *dude;
+	uid_t euid;
+	gid_t egid;
+	
+	euid = geteuid();
+	egid = getegid();
+	
+	if (strncmp("cf",file,2)) {
+		if ((dude=getpwnam(euser)) == (struct passwd *)0 || !*euser)
+		  return(ERROR);
+	
+		if (setegid(dude->pw_gid) < 0)
+		  return (ERROR);
+		if (seteuid(dude->pw_uid) < 0)
+		  return (ERROR);
+	}
+	
+	/* Open the file as the user. */
+
+	if (lstat(file, &stb) < 0 || (f = open(file, O_RDONLY)) < 0) {
+		seteuid(euid);
+		setegid(egid);
+		return(ACCESS);
+	}
+	
+	if (seteuid(euid) < 0 || setegid(egid) < 0) {
+		syslog(LOG_ERR, "can't reset effective uid/gid!");
+		return (ERROR);
+	}
 
-	if (lstat(file, &stb) < 0 || (f = open(file, O_RDONLY)) < 0)
-		return(ERROR);
 	/*
 	 * Check to see if data file is a symbolic link. If so, it should
 	 * still point to the same file or someone is trying to print something
@@ -1184,6 +1328,21 @@ unlinkfile(file, host, user)
 	char *host;
 	char *user;
 {
+	/* this hack duplicates some code but things should work better */
+	/* if the file does not exist, do not worry about doing unlink */
+	int ret;
+	extern int errno;
+	if ((strchr(file,'/') == NULL) &&
+	    (strstr(file,"..") == NULL)) {
+		ret = unlink(file);
+		if (ret == 0)
+			return;
+		if (errno == 2)
+			return;
+		syslog(LOG_ERR,"unlink file=%s, ret=%i, errno=%i",
+			file, ret, errno);
+	}
+	
 	if (dounlink(file, host, user))
 		sendmail(user, UNLINK);
 }
@@ -1200,7 +1359,6 @@ sendmail(user, bombed)
 	register int i;
 	int p[2], s;
 	register char *cp;
-	char buf[100];
 	struct stat stb;
 	FILE *fp;
 
@@ -1213,8 +1371,7 @@ sendmail(user, bombed)
 			cp++;
 		else
 			cp = _PATH_SENDMAIL;
-		sprintf(buf, "%s@%s", user, fromhost);
-		execl(_PATH_SENDMAIL, cp, buf, 0);
+		execl(_PATH_SENDMAIL, cp, "-t", 0);
 		exit(0);
 	} else if (s > 0) {				/* parent */
 		dup2(p[1], 1);
@@ -1314,6 +1471,7 @@ doforkuser(action, user)
 {
 	register int i, pid;
 	struct passwd *dude;
+	struct group *lpgrp;
 	
 	/* figure out who we wish we were */
 	dude = getpwnam(user);
@@ -1331,6 +1489,20 @@ doforkuser(action, user)
 	  }
 	}
 	
+	/* This shouldn't fail */
+	lpgrp = getgrnam("lp");
+	if (!lpgrp) {
+		syslog(LOG_ERR, "can't find lp group - can't fork");
+		switch (action) {
+		 case DORETURN:
+			return (-1);
+		 default:
+			syslog(LOG_ERR, "bad action (%d) to dofork", action);
+		 case DOABORT:
+			exit(1);
+		}
+	}
+	
 	for (i = 0; i < 20; i++) {
 		if ((pid = fork()) < 0) {
 			sleep((unsigned)(i*i));
@@ -1338,10 +1510,15 @@ doforkuser(action, user)
 		}
 		/*
 		 * Child should run as user instead of root
+		 * Add to lp group so it can read config file.
 		 */
 		if (pid == 0) {
-		   setuid(dude->pw_uid);
-		   setgid(dude->pw_gid);
+			unsetenv("BASH_ENV");
+			unsetenv("ENV");
+			setenv("USER",user,1);
+			setgroups(1,&lpgrp->gr_gid);
+			setgid(dude->pw_gid);
+			setuid(dude->pw_uid);
 		}
 		return(pid);
 	}
@@ -1464,6 +1641,44 @@ openpr()
 	char *cp;
 
 	if (!sendtorem && *LP) {
+               if ((cp = strchr(LP, '@')) != NULL) {
+                       char savech;
+                       struct servent *sp;
+                       struct hostent *hp;
+                       struct sockaddr_in sin;
+
+                       savech = *cp;
+                       *cp = '\0';
+                       if ((sp = getservbyname(LP, "tcp")) != NULL) {
+                               sin.sin_port = sp->s_port;
+                       } else if (isdigit(*LP)) {
+                               sin.sin_port = htons(atoi(LP));
+                       } else {
+                               syslog(LOG_ERR, "%s: bad service spec", LP);
+                               exit(1);
+                       }
+                       *cp = savech;
+                       if (inet_aton(cp+1, &sin.sin_addr))
+                               NULL;
+                       else if ((hp = gethostbyname(cp+1)) != NULL)
+                               sin.sin_addr.s_addr = *(u_long *)hp->h_addr;
+                       else {
+                               syslog(LOG_ERR, "%s: bad host spec", cp+1);
+                               exit(1);
+                       }
+                       if ((pfd = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) <0){
+                               syslog(LOG_ERR, "socket: %m");
+                               exit(1);
+                       }
+#ifndef __linux__
+                       sin.sin_len = sizeof sin;
+#endif
+                       sin.sin_family = AF_INET;
+                       if (connect(pfd, (struct sockaddr*)&sin, sizeof sin) <0){
+                               syslog(LOG_ERR, "connect: %m");
+                               exit(1);
+                       }
+               } else {
 		for (i = 1; ; i = i < 32 ? i << 1 : i) {
 			pfd = open(LP, RW ? O_RDWR : O_WRONLY);
 			if (pfd >= 0)
@@ -1478,17 +1693,26 @@ openpr()
 		}
 		if (isatty(pfd))
 			setty();
+		}
 		pstatus("%s is ready and printing", printer);
 	} else if (RM != NULL) {
 		for (i = 1; ; i = i < 256 ? i << 1 : i) {
 			resp = -1;
 			pfd = getport(RM);
 			if (pfd >= 0) {
+				char c;
+				
 				(void) snprintf(line, sizeof(line), "\2%s\n", RP);
 				n = strlen(line);
 				if (write(pfd, line, n) == n &&
 				    (resp = response()) == '\0')
 					break;
+#ifdef SHUT_WR
+				shutdown(pfd, SHUT_WR);
+#else
+				shutdown(pfd, 1);
+#endif
+				read(pfd , &c, 1);
 				(void) close(pfd);
 			}
 			if (i == 1) {
