@@ -60,6 +60,8 @@ char copyright[] =
 #include "lp.local.h"
 #include "pathnames.h"
 
+#define MAXHOSTNAMELEN 128
+
 char    *tfname;		/* tmp copy of cf before linking */
 char    *cfname;		/* daemon control files, linked from tf's */
 char    *dfname;		/* data files */
@@ -81,7 +83,7 @@ char	*person;		/* user name */
 char	*title;			/* pr'ing title */
 char	*fonts[4];		/* troff font names */
 char	*width;			/* width for versatec printing */
-char	host[32];		/* host name */
+char	host[MAXHOSTNAMELEN];		/* host name */
 char	*class = host;		/* class title on header page */
 char    *jobname;		/* job name on header page */
 char	*name;			/* program name */
@@ -96,6 +98,9 @@ char	*LO;			/* lock file name */
 char	*RG;			/* restrict group */
 static short	SC;			/* suppress multiple copies */
 
+uid_t	euid,uid;
+gid_t	egid,gid;
+
 char	*getenv();
 char	*rindex();
 char	*linked();
@@ -113,6 +118,13 @@ main(argc, argv)
 	char buf[BUFSIZ];
 	int i, f;
 	struct stat stb;
+	
+	euid = geteuid();
+	egid = getegid();
+	gid = getgid();
+	uid = getuid();
+	setegid(gid);
+	seteuid(uid);
 
 	if (signal(SIGHUP, SIG_IGN) != SIG_IGN)
 		signal(SIGHUP, cleanup);
@@ -125,6 +137,7 @@ main(argc, argv)
 
 	name = argv[0];
 	gethostname(host, sizeof (host));
+	host[MAXHOSTNAMELEN-1]='\0';
 	openlog("lpd", 0, LOG_LPR);
 
 	while (argc > 1 && argv[1][0] == '-') {
@@ -253,7 +266,7 @@ main(argc, argv)
 	 * Get the identity of the person doing the lpr using the same
 	 * algorithm as lprm. 
 	 */
-	userid = getuid();
+	userid = uid;
 	if (userid != DU || person == 0) {
 		if ((pw = getpwuid(userid)) == NULL)
 			fatal("Who are you?");
@@ -265,7 +278,7 @@ main(argc, argv)
 	if (RG != NULL && userid != DU) {
 		if ((gptr = getgrnam(RG)) == NULL)
 			fatal("Restricted group specified incorrectly");
-		if (gptr->gr_gid != getgid()) {
+		if (gptr->gr_gid != gid) {
 			while (*gptr->gr_mem != NULL) {
 				if ((strcmp(person, *gptr->gr_mem)) == 0)
 					break;
@@ -286,7 +299,11 @@ main(argc, argv)
 	 */
 	mktemps();
 	tfd = nfile(tfname);
+	seteuid(euid);
+	setegid(egid);
 	(void) fchown(tfd, DU, -1);	/* owned by daemon for protection */
+	setegid(gid);
+	seteuid(uid);
 	card('H', host);
 	card('P', person);
 	if (hdr) {
@@ -321,7 +338,7 @@ main(argc, argv)
 			continue;	/* file unreasonable */
 
 		if (sflag && (cp = linked(arg)) != NULL) {
-			(void) sprintf(buf, "%ld %ld",
+			(void) snprintf(buf, sizeof(buf), "%ld %ld",
 				   (long) statb.st_dev, (long) statb.st_ino);
 			card('S', buf);
 			if (format == 'p')
@@ -338,14 +355,15 @@ main(argc, argv)
 		}
 		if (sflag)
 			printf("%s: %s: not linked, copying instead\n", name, arg);
+		
 		if ((i = open(arg, O_RDONLY)) < 0) {
 			printf("%s: cannot open %s\n", name, arg);
-			continue;
+		} else {
+			copy(i, arg);
+			(void) close(i);
+			if (f && unlink(arg) < 0)
+			  printf("%s: %s: not removed\n", name, arg);
 		}
-		copy(i, arg);
-		(void) close(i);
-		if (f && unlinkfile(arg) < 0)
-			printf("%s: %s: not removed\n", name, arg);
 	}
 
 	if (nact) {
@@ -354,6 +372,8 @@ main(argc, argv)
 		/*
 		 * Touch the control file to fix position in the queue.
 		 */
+		seteuid(euid);
+		setegid(egid);
 		if ((tfd = open(tfname, O_RDWR)) >= 0) {
 			char c;
 
@@ -375,6 +395,8 @@ main(argc, argv)
 			exit(0);
 		if (!startdaemon(printer))
 			printf("jobs queued, but cannot start daemon.\n");
+		setegid(gid);
+		seteuid(uid);
 		exit(0);
 	}
 	cleanup();
@@ -431,6 +453,7 @@ linked(file)
 {
 	register char *cp;
 	static char buf[BUFSIZ];
+	int ret;
 
 	if (*file != '/') {
 		if (getcwd(buf, BUFSIZ) == NULL)
@@ -454,7 +477,12 @@ linked(file)
 		strncat(buf, file, sizeof(buf) - strlen(buf) - 1);
 		file = buf;
 	}
-	return(symlink(file, dfname) ? NULL : file);
+	seteuid(euid);
+	setegid(egid);
+	ret=symlink(file, dfname);
+	setegid(gid);
+	seteuid(uid);
+	return(ret ? NULL : file);
 }
 
 /*
@@ -490,7 +518,9 @@ nfile(n)
 {
 	register f;
 	int oldumask = umask(0);		/* should block signals */
-
+	
+	seteuid(euid);
+	setegid(egid);
 	f = creat(n, FILMOD);
 	(void) umask(oldumask);
 	if (f < 0) {
@@ -501,6 +531,8 @@ nfile(n)
 		printf("%s: cannot chown %s\n", name, n);
 		cleanup();
 	}
+	setegid(gid);
+	seteuid(uid);
 	if (++n[inchar] > 'z') {
 		if (++n[inchar-2] == 't') {
 			printf("too many files - break up the job\n");
@@ -525,6 +557,8 @@ cleanup()
 	signal(SIGQUIT, SIG_IGN);
 	signal(SIGTERM, SIG_IGN);
 	i = inchar;
+	seteuid(euid);
+	setegid(egid);
 	if (tfname)
 		do
 			unlink(tfname);
@@ -553,7 +587,7 @@ test(file)
 {
 	register int fd;
 	register char *cp;
-
+	
 	if (access(file, 4) < 0) {
 		printf("%s: cannot access %s\n", name, file);
 		return(-1);
@@ -600,29 +634,6 @@ error1:
 	return(-1);
 }
 
-int
-unlinkfile(file)
-	register char *file;
-{
-	uid_t	euid = geteuid();
-	gid_t	egid = getegid();
-	int	err;
-
-	if (setegid(getgid()) < 0)
-		return 0;
-	if (seteuid(getuid()) < 0)
-		return 0;
-	if ((err = unlink(file)) < 0) {
-		printf("%s: %s: couldn't remove file: %s\n",
-				name, file, strerror(errno));
-	}
-	if (seteuid(euid) < 0 || setegid(egid) < 0) {
-		printf("%s: couldn't reset effective uid/gid!\n", name);
-		cleanup();
-	}
-	return (err >= 0);
-}
-
 /*
  * itoa - integer to string conversion
  */
@@ -632,6 +643,9 @@ itoa(i)
 {
 	static char b[10] = "########";
 	register char *p;
+	
+	if ((i<0) || (i > 99999999))
+	  i=0;
 
 	p = &b[8];
 	do
@@ -680,6 +694,8 @@ mktemps()
 	char buf[BUFSIZ];
 
 	(void) snprintf(buf, sizeof(buf), "%s/.seq", SD);
+	seteuid(euid);
+	setegid(egid);
 	if ((fd = open(buf, O_RDWR|O_CREAT, 0661)) < 0) {
 		printf("%s: cannot create %s\n", name, buf);
 		exit(1);
@@ -688,6 +704,8 @@ mktemps()
 		printf("%s: cannot lock %s\n", name, buf);
 		exit(1);
 	}
+	setegid(gid);
+	seteuid(uid);
 	n = 0;
 	if ((len = read(fd, buf, sizeof(buf))) > 0) {
 		for (cp = buf; len--; ) {
@@ -696,12 +714,10 @@ mktemps()
 			n = n * 10 + (*cp++ - '0');
 		}
 	}
-	len = strlen(SD) + 15;
+	len = strlen(SD) + strlen(host) +8;
 	if ((cfname = malloc(len)) == NULL)
 		fatal("out of memory");
-	(void) sprintf(cfname, "%s/cfA%03dAXXXXXX", SD, n);
-	if (mktemp(cfname) == NULL)
-		fatal("couldn't create unique control file");
+	(void) snprintf(cfname, len, "%s/cfA%03d%s", SD, n, host);
 	if ((tfname = strdup(cfname)) == NULL || (dfname = strdup(cfname)) == NULL)
 		fatal("out of memory");
 	inchar = strlen(SD) + 3;
@@ -709,7 +725,7 @@ mktemps()
 	dfname[inchar-2] = 'd';
 	n = (n + 1) % 1000;
 	(void) lseek(fd, 0L, 0);
-	sprintf(buf, "%03d\n", n);
+	snprintf(buf, sizeof(buf), "%03d\n", n);
 	(void) write(fd, buf, strlen(buf));
 	(void) close(fd);	/* unlocks as well */
 }
